//! Utilities for processing Mercurial command-server protocol.

use bytes::{Buf, BufMut, Bytes, BytesMut};
use std::io::{self, Cursor};
use std::mem;
use tokio_util::codec::{Decoder, Encoder};

/// Message sent from Mercurial command server.
#[derive(Clone, Debug, Eq, PartialEq)]
pub enum ChannelMessage {
    /// Data sent from server to the specified channel.
    Data(u8, Bytes),
    /// Server requesting data up to the specified size in bytes.
    InputRequest(usize),
    /// Server requesting a single-line input up to the specified size in bytes.
    LineRequest(usize),
    /// Server requesting a shell command execution at client side. (cHg extension)
    SystemRequest(Bytes),
}

/// Decoder to parse and split messages sent from Mercurial command server.
#[derive(Debug)]
pub struct ChannelDecoder {}

impl ChannelDecoder {
    pub fn new() -> ChannelDecoder {
        ChannelDecoder {}
    }
}

impl Decoder for ChannelDecoder {
    type Item = ChannelMessage;
    type Error = io::Error; // TODO: maybe introduce a dedicated error type

    fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>, Self::Error> {
        parse_channel_header(src.as_ref())
            .map_or(Ok(None), |(ch, len)| decode_channel_payload(src, ch, len))
    }
}

const CHANNEL_HEADER_LEN: usize = mem::size_of::<u8>() + mem::size_of::<u32>();

fn parse_channel_header(src: &[u8]) -> Option<(u8, usize)> {
    if src.len() < CHANNEL_HEADER_LEN {
        return None;
    }
    let mut buf = Cursor::new(src);
    let ch = buf.get_u8();
    let len = buf.get_u32();
    Some((ch, len as usize))
}

fn decode_channel_payload(
    src: &mut BytesMut,
    ch: u8,
    payload_len: usize,
) -> Result<Option<ChannelMessage>, io::Error> {
    let has_payload = ch.is_ascii_lowercase() || ch == b'S';
    if has_payload && src.len() < CHANNEL_HEADER_LEN + payload_len {
        return Ok(None);
    }
    src.advance(CHANNEL_HEADER_LEN);
    match ch {
        b'I' => Ok(Some(ChannelMessage::InputRequest(payload_len))),
        b'L' => Ok(Some(ChannelMessage::LineRequest(payload_len))),
        b'S' => {
            let payload = src.split_to(payload_len).freeze();
            Ok(Some(ChannelMessage::SystemRequest(payload)))
        }
        c if has_payload => {
            let payload = src.split_to(payload_len).freeze();
            Ok(Some(ChannelMessage::Data(c, payload)))
        }
        c => {
            let msg = format!("unknown required channel: {}", c);
            Err(io::Error::new(io::ErrorKind::InvalidData, msg))
        }
    }
}

/// Request and response sent to Mercurial command server.
#[derive(Clone, Debug, Eq, PartialEq)]
pub enum BlockMessage {
    /// Command to initiate new request, without argument.
    Command(Bytes),
    /// Command argument or data sent to server.
    Data(Bytes),
}

/// Encoder to build messages sent to Mercurial command server.
#[derive(Debug)]
pub struct BlockEncoder {}

impl BlockEncoder {
    pub fn new() -> BlockEncoder {
        BlockEncoder {}
    }
}

impl Encoder<BlockMessage> for BlockEncoder {
    type Error = io::Error; // TODO: maybe introduce a dedicated error type

    fn encode(&mut self, msg: BlockMessage, dst: &mut BytesMut) -> Result<(), Self::Error> {
        match msg {
            BlockMessage::Command(cmd) => encode_command_line(cmd, dst),
            BlockMessage::Data(data) => encode_data_block(data, dst),
        }
    }
}

fn encode_command_line(cmd: Bytes, dst: &mut BytesMut) -> Result<(), io::Error> {
    // TODO: error out if cmd contains '\n'?
    dst.reserve(cmd.len() + 1);
    dst.put(cmd);
    dst.put_u8(b'\n');
    Ok(())
}

fn encode_data_block(data: Bytes, dst: &mut BytesMut) -> Result<(), io::Error> {
    if data.len() > u32::max_value() as usize {
        let msg = format!("data length exceeds protocol limit: {}", data.len());
        return Err(io::Error::new(io::ErrorKind::InvalidInput, msg));
    }
    dst.reserve(mem::size_of::<u32>() + data.len());
    dst.put_u32(data.len() as u32);
    dst.put(data);
    Ok(())
}

/// Unified codec for client-side operations.
///
/// With this codec, a framed stream can be constructed without splitting
/// it into read/write halves.
#[derive(Debug)]
pub struct ClientCodec {
    dec: ChannelDecoder,
    enc: BlockEncoder,
}

impl ClientCodec {
    pub fn new() -> ClientCodec {
        ClientCodec {
            dec: ChannelDecoder::new(),
            enc: BlockEncoder::new(),
        }
    }
}

impl Decoder for ClientCodec {
    type Item = ChannelMessage;
    type Error = io::Error;

    fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>, Self::Error> {
        self.dec.decode(src)
    }
}

impl Encoder<BlockMessage> for ClientCodec {
    type Error = io::Error;

    fn encode(&mut self, msg: BlockMessage, dst: &mut BytesMut) -> Result<(), Self::Error> {
        self.enc.encode(msg, dst)
    }
}
