//! Abstraction to support both socket-like streams and pipes.

use futures::{Sink, Stream};
use std::io;
use tokio::process::{Child, ChildStdin, ChildStdout};
use tokio_util::codec::{FramedRead, FramedWrite};

use crate::codec::{BlockEncoder, BlockMessage, ChannelDecoder, ChannelMessage};

/// Leaky abstraction to hold connection to a command server.
pub trait Connection {
    type Rx: Stream<Item = io::Result<ChannelMessage>> + Unpin;
    type Tx: Sink<BlockMessage, Error = io::Error> + Unpin;

    fn get_rx_mut(&mut self) -> &mut Self::Rx;
    fn get_tx_mut(&mut self) -> &mut Self::Tx;
}

/// Wrapper for `Child` process to keep it alive and provide stdio channels.
#[derive(Debug)]
pub struct PipeConnection {
    rx: FramedRead<ChildStdout, ChannelDecoder>,
    tx: FramedWrite<ChildStdin, BlockEncoder>,
    child: Child,
}

impl PipeConnection {
    pub fn new(mut child: Child) -> PipeConnection {
        let stdout = child.stdout.take().unwrap();
        let stdin = child.stdin.take().unwrap();
        let rx = FramedRead::new(stdout, ChannelDecoder::new());
        let tx = FramedWrite::new(stdin, BlockEncoder::new());
        PipeConnection { rx, tx, child }
    }
}

impl Connection for PipeConnection {
    type Rx = FramedRead<ChildStdout, ChannelDecoder>;
    type Tx = FramedWrite<ChildStdin, BlockEncoder>;

    fn get_rx_mut(&mut self) -> &mut Self::Rx {
        &mut self.rx
    }

    fn get_tx_mut(&mut self) -> &mut Self::Tx {
        &mut self.tx
    }
}

#[cfg(unix)]
pub use self::unix::UnixConnection;

#[cfg(unix)]
mod unix {
    use std::os::unix::io::{AsRawFd, RawFd};
    use tokio::net::UnixStream;
    use tokio_util::codec::{Decoder, Framed};

    use super::*;
    use crate::codec::ClientCodec;

    /// Wrapper for `UnixStream` which has no separate read/write channels.
    #[derive(Debug)]
    pub struct UnixConnection {
        framed: Framed<UnixStream, ClientCodec>,
    }

    impl UnixConnection {
        pub fn new(stream: UnixStream) -> UnixConnection {
            let framed = ClientCodec::new().framed(stream);
            UnixConnection { framed }
        }
    }

    impl Connection for UnixConnection {
        type Rx = Framed<UnixStream, ClientCodec>;
        type Tx = Framed<UnixStream, ClientCodec>;

        fn get_rx_mut(&mut self) -> &mut Self::Rx {
            &mut self.framed
        }

        fn get_tx_mut(&mut self) -> &mut Self::Tx {
            &mut self.framed
        }
    }

    impl AsRawFd for UnixConnection {
        fn as_raw_fd(&self) -> RawFd {
            self.framed.get_ref().as_raw_fd()
        }
    }
}
