//! Utilities for parsing and building command-server messages.

use bytes::{Buf, BufMut, Bytes, BytesMut};
use std::cmp;
use std::collections::HashSet;
use std::error;
#[cfg(unix)]
use std::ffi::OsStr;
use std::io::{self, Cursor};
use std::mem;
#[cfg(unix)]
use std::os::unix::ffi::OsStrExt;

/// Information reported by the server at the initial handshake.
#[derive(Clone, Debug, Eq, PartialEq)]
pub struct ServerSpec {
    pub capabilities: HashSet<String>,
    pub encoding: Option<String>,
    pub process_id: Option<u32>,
    pub process_group_id: Option<u32>,
}

/// Parses "hello" response into `ServerSpec`.
pub fn parse_hello(src: Bytes) -> io::Result<ServerSpec> {
    let mut capabilities = HashSet::new();
    let mut encoding = None;
    let mut process_id = None;
    let mut process_group_id = None;

    for line in decode_latin1(src).lines() {
        let mut split = line.splitn(2, ':').fuse();
        let key = split.next().unwrap();
        let value = split
            .next()
            .map(|s| s.trim_start())
            .ok_or(new_parse_error(format!(
                "malformed line in hello message: {}",
                line
            )))?;

        match key {
            "capabilities" => {
                capabilities.extend(value.split_whitespace().map(|s| s.to_owned()));
            }
            "encoding" => {
                encoding = Some(value.to_owned());
            }
            "pid" => {
                let n = value
                    .parse()
                    .map_err(|_| new_parse_error(format!("malformed pid: {}", value)))?;
                process_id = Some(n);
            }
            "pgid" => {
                let n = value
                    .parse()
                    .map_err(|_| new_parse_error(format!("malformed pgid: {}", value)))?;
                process_group_id = Some(n);
            }
            _ => {} // ignores unknown key
        }
    }

    Ok(ServerSpec {
        capabilities,
        encoding,
        process_id,
        process_group_id,
    })
}

/// Parses command exit code into integer.
pub fn parse_result_code(data: Bytes) -> io::Result<i32> {
    if data.len() < mem::size_of::<i32>() {
        let msg = format!("result code too short: {}", data.len());
        return Err(io::Error::new(io::ErrorKind::InvalidData, msg));
    }
    Ok(Cursor::new(data).get_i32())
}

/// Packs command exit code into bytes.
pub fn pack_result_code(code: i32) -> Bytes {
    let mut data = BytesMut::with_capacity(mem::size_of::<i32>());
    data.put_i32(code);
    data.freeze()
}

/// Packs command arguments of arbitrary encoding into bytes.
///
/// # Panics
///
/// Panics if argument contains `\0` character.
pub fn pack_args(args: impl IntoIterator<Item = impl AsRef<[u8]>>) -> Bytes {
    let mut args_iter = args.into_iter();
    if let Some(a) = args_iter.next() {
        assert!(!a.as_ref().contains(&0), "argument shouldn't contain NUL");
        let mut dst = BytesMut::with_capacity(cmp::max(a.as_ref().len(), 200));
        dst.put_slice(a.as_ref());
        for a in args_iter {
            assert!(!a.as_ref().contains(&0), "argument shouldn't contain NUL");
            dst.reserve(1 + a.as_ref().len());
            dst.put_u8(b'\0');
            dst.put_slice(a.as_ref());
        }
        dst.freeze()
    } else {
        Bytes::new()
    }
}

/// Packs command arguments of platform encoding into bytes.
#[cfg(unix)]
pub fn pack_args_os(args: impl IntoIterator<Item = impl AsRef<OsStr>>) -> Bytes {
    pack_args(args.into_iter().map(|a| a.as_ref().as_bytes().to_owned()))
}

fn decode_latin1(s: impl AsRef<[u8]>) -> String {
    s.as_ref().iter().map(|&c| c as char).collect()
}

// TODO: error type
fn new_parse_error(error: impl Into<Box<dyn error::Error + Send + Sync>>) -> io::Error {
    io::Error::new(io::ErrorKind::InvalidData, error)
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::iter::FromIterator;
    use std::panic;

    #[test]
    fn parse_hello_empty() {
        let spec = ServerSpec {
            capabilities: HashSet::new(),
            encoding: None,
            process_id: None,
            process_group_id: None,
        };
        assert_eq!(parse_hello(Bytes::from_static(b"")).unwrap(), spec);
    }

    #[test]
    fn parse_hello_some() {
        let src = [
            b"capabilities: getencoding runcommand".as_ref(),
            b"encoding: utf-8".as_ref(),
            b"pid: 12345".as_ref(),
            b"pgid: 6789",
        ]
        .join(&b'\n');
        let spec = ServerSpec {
            capabilities: HashSet::from_iter(
                ["getencoding", "runcommand"].iter().map(|&s| s.to_owned()),
            ),
            encoding: Some("utf-8".to_owned()),
            process_id: Some(12345),
            process_group_id: Some(6789),
        };
        assert_eq!(parse_hello(Bytes::from(src)).unwrap(), spec);
    }

    #[test]
    fn parse_hello_unsupported() {
        let src = [
            b"capabilities: runcommand unsupported".as_ref(),
            b"unsupported: value".as_ref(),
            b"pid: 12345".as_ref(),
        ]
        .join(&b'\n');
        let spec = ServerSpec {
            capabilities: HashSet::from_iter(
                ["runcommand", "unsupported"].iter().map(|&s| s.to_owned()),
            ),
            encoding: None,
            process_id: Some(12345),
            process_group_id: None,
        };
        assert_eq!(parse_hello(Bytes::from(src)).unwrap(), spec);
    }

    #[test]
    fn parse_hello_malformed() {
        assert_eq!(
            parse_hello(Bytes::from_static(b"\n")).unwrap_err().kind(),
            io::ErrorKind::InvalidData
        );
        assert_eq!(
            parse_hello(Bytes::from_static(b"caba")).unwrap_err().kind(),
            io::ErrorKind::InvalidData
        );
        assert_eq!(
            parse_hello(Bytes::from_static(b"pid: foo"))
                .unwrap_err()
                .kind(),
            io::ErrorKind::InvalidData
        );
        assert_eq!(
            parse_hello(Bytes::from_static(b"pgid: foo"))
                .unwrap_err()
                .kind(),
            io::ErrorKind::InvalidData
        );
    }

    #[test]
    fn pack_args_some() {
        assert_eq!(pack_args(&[] as &[&[u8]]), Bytes::new());
        assert_eq!(pack_args(&[b"foo".as_ref()]), Bytes::from_static(b"foo"));
        assert_eq!(
            pack_args(&[b"foo".as_ref(), b"".as_ref(), b"bar".as_ref()]),
            Bytes::from_static(b"foo\0\0bar")
        );
    }

    #[test]
    fn pack_args_nul() {
        assert!(panic::catch_unwind(|| pack_args(&[b"\0"])).is_err());
        assert!(panic::catch_unwind(|| pack_args(&[b"fo\0"])).is_err());
        assert!(panic::catch_unwind(|| pack_args(&[b"foo", b"\0ar"])).is_err());
        assert!(panic::catch_unwind(|| pack_args(&[b"foo", b"bar", b"b\0z"])).is_err());
    }
}
