//! Low-level building blocks for the command server protocol.

use bytes::Bytes;
use futures::SinkExt;
use std::io;
use tokio_stream::{self as stream, StreamExt};

use crate::codec::{BlockMessage, ChannelMessage};
use crate::connection::Connection;

/// Connection wrapper to process channel requests and responses.
///
/// This provides a low-level interface to communicate with the Mercurial
/// command server.
#[derive(Debug)]
pub struct Protocol<C> {
    conn: C,
}

impl<C> Protocol<C> {
    /// Creates a new Protocol that wraps the given connection.
    pub fn new(conn: C) -> Self {
        Protocol { conn }
    }

    /// Unwraps the underlying `Connection`.
    pub fn into_connection(self) -> C {
        self.conn
    }
}

#[cfg(unix)]
mod unix {
    use super::*;
    use std::os::unix::io::{AsRawFd, RawFd};

    impl<C> AsRawFd for Protocol<C>
    where
        C: AsRawFd,
    {
        fn as_raw_fd(&self) -> RawFd {
            self.conn.as_raw_fd()
        }
    }
}

impl<C> Protocol<C>
where
    C: Connection,
{
    /// Sends the command of no argument without waiting for response.
    ///
    /// This is equivalent to `OneShotRequest::start()` of tokio-hglib 0.2.
    /// For `MessageLoop::start()`, call this function and fetch responses.
    pub async fn send_command(&mut self, cmd: impl Into<Bytes>) -> io::Result<()> {
        self.conn
            .get_tx_mut()
            .send(BlockMessage::Command(cmd.into()))
            .await?;
        Ok(())
    }

    /// Sends the command and arguments without waiting for response.
    ///
    /// This is equivalent to `OneShotRequest::start_with_args()` of tokio-hglib 0.2.
    /// For `MessageLoop::start_with_args()`, call this function and fetch responses.
    pub async fn send_command_with_args(
        &mut self,
        cmd: impl Into<Bytes>,
        packed_args: impl Into<Bytes>,
    ) -> io::Result<()> {
        let blocks = vec![
            Ok(BlockMessage::Command(cmd.into())),
            Ok(BlockMessage::Data(packed_args.into())),
        ];
        self.conn
            .get_tx_mut()
            .send_all(&mut stream::iter(blocks))
            .await?;
        Ok(())
    }

    /// Sends the given data back to the server.
    ///
    /// For `MessageLoop::resume_with_data()` of tokio-hglib 0.2, call this function
    /// and fetch responses.
    pub async fn send_data(&mut self, data: impl Into<Bytes>) -> io::Result<()> {
        self.conn
            .get_tx_mut()
            .send(BlockMessage::Data(data.into()))
            .await?;
        Ok(())
    }

    /// Sends the command of no argument, and fetches the result data.
    ///
    /// This is equivalent to `OneShotQuery::start()` of tokio-hglib 0.2.
    pub async fn query(&mut self, cmd: impl Into<Bytes>) -> io::Result<Bytes> {
        self.send_command(cmd).await?;
        self.fetch_result().await
    }

    /// Sends the command and arguments, and fetches the result data.
    ///
    /// This is equivalent to `OneShotQuery::start_with_args()` of tokio-hglib 0.2.
    pub async fn query_with_args(
        &mut self,
        cmd: impl Into<Bytes>,
        packed_args: impl Into<Bytes>,
    ) -> io::Result<Bytes> {
        self.send_command_with_args(cmd, packed_args).await?;
        self.fetch_result().await
    }

    /// Fetches response message from the server.
    ///
    /// This is equivalent to `MessageLoop::resume()` of tokio-hglib 0.2.
    pub async fn fetch_response(&mut self) -> io::Result<ChannelMessage> {
        let v = self.conn.get_rx_mut().try_next().await?;
        expect_msg(v)
    }

    async fn fetch_result(&mut self) -> io::Result<Bytes> {
        loop {
            match self.fetch_response().await? {
                ChannelMessage::Data(b'r', data) => {
                    return Ok(data);
                }
                ChannelMessage::Data(..) => {
                    // just ignore data sent to uninteresting (optional) channel
                }
                ChannelMessage::InputRequest(..)
                | ChannelMessage::LineRequest(..)
                | ChannelMessage::SystemRequest(..) => {
                    return Err(io::Error::new(
                        io::ErrorKind::InvalidData,
                        "unsupported request while querying",
                    ));
                }
            }
        }
    }
}

fn expect_msg(v: Option<ChannelMessage>) -> Result<ChannelMessage, io::Error> {
    v.ok_or(io::Error::new(
        io::ErrorKind::UnexpectedEof,
        "no result code received",
    ))
}
