//! Low-level functions to run Mercurial command in command server.

use async_trait::async_trait;
use bytes::Bytes;
use std::io;

use crate::codec::ChannelMessage;
use crate::connection::Connection;
use crate::message;
use crate::protocol::Protocol;

/// Callback to process data and requests received from the server.
#[async_trait]
pub trait UiHandler {
    /// Handles bulk input request.
    async fn read_data(&mut self, len: usize) -> io::Result<Bytes>;

    /// Handles line input request.
    async fn read_line(&mut self, len: usize) -> io::Result<Bytes>;

    /// Handles data received through the output channel.
    async fn write_output(&mut self, data: Bytes) -> io::Result<()>;

    /// Handles data received through the error channel.
    async fn write_error(&mut self, data: Bytes) -> io::Result<()>;
}

/// Runs the given Mercurial command, and fetches the result code.
pub async fn run_command(
    proto: &mut Protocol<impl Connection>,
    handler: &mut impl UiHandler,
    packed_args: impl Into<Bytes>,
) -> io::Result<i32> {
    proto
        .send_command_with_args("runcommand", packed_args)
        .await?;
    loop {
        match proto.fetch_response().await? {
            ChannelMessage::Data(b'r', data) => {
                return message::parse_result_code(data);
            }
            ChannelMessage::Data(b'o', data) => {
                handler.write_output(data).await?;
            }
            ChannelMessage::Data(b'e', data) => {
                handler.write_error(data).await?;
            }
            ChannelMessage::Data(..) => {
                // just ignores data sent to unknown (optional) channel
            }
            ChannelMessage::InputRequest(len) => {
                let data = handler.read_data(len).await?;
                proto.send_data(data).await?;
            }
            ChannelMessage::LineRequest(len) => {
                let data = handler.read_line(len).await?;
                proto.send_data(data).await?;
            }
            ChannelMessage::SystemRequest(..) => {
                return Err(io::Error::new(
                    io::ErrorKind::InvalidData,
                    "unsupported request",
                ));
            }
        }
    }
}
